#!/usr/bin/env python

import sys
import time

import auxlib

# Sample code for communicating with the PlaneWave mirror cover controller.
# The controller uses the same "aux" protocol that is used by the EFA focuser/rotator
# accessory, so the efalib module is used to form the requets and parse the responses.

# This code is compatible with Python 2.7 or Python 3.x.

# This can be called from the command line with an argument, as follows:
#   python pwshutter.py open
#   python pwshutter.py close
#   python pwshutter.py stop
#   python pwshutter.py reset
#   python pwshutter.py status
#
# Or it can be imported from another Python script and called programmatically. 
# For example:
#  import pwshutter
#  pwshutter.connect("/dev/ttyUSB0")
#  pwshutter.open()
#  pwshutter.status()


# TODO: Change these settings as needed if calling this program from the command line
CONNECT_VIA_TCP = False  # If False, use a serial port. If True, connect via TCP/IP (over the Ethernet cable)
#DEFAULT_SERIAL_PORT = "/dev/ttyACM0"  # Sample for Linux
DEFAULT_SERIAL_PORT = "COM18"  # Sample for Windows
#DEFAULT_TCP_HOST = "10.17.249.37"
#DEFAULT_TCP_PORT = 2242

DEBUG_MESSAGES = False  # If True, this will log additional debugging info, such as the bytes being exchanged with the controller


ADDR_SHUTTER = 0x17 # Aux device address for the main shutter control box

# Flapper commands
FLM_OPEN =   0x01    # Starts open sequence
FLM_CLOSE =  0x02    # Starts close sequence
FLM_SPEED =  0x03    # Sets/Gets speed for flappers -- NOT IMPLEMENTED
FLM_STATUS = 0x04    # Gets flapper status 
FLM_STOP =   0x05    # Stops open/close sequence
FLM_ADC =    0x06    # returns ADC of each flapper -- NOT IMPLEMENTED
FLM_RESET =  0x0F    # resets the flapper state
FLM_GET_VERSION = 0xFE # Get firmware version

# FLM_STATUS response byte 0: Flapper Master state
FLM_STATE_UNKNOWN = 0 # flapper needs update -- not all flappers have reported
FLM_STATE_OPEN =    1 # flapper is open
FLM_STATE_CLOSED =  2 # flapper is closed
FLM_STATE_BETWEEN = 3 # flapper is between switches
FLM_STATE_ERR =     4 # flapper is in fault -- over current / temp on one or more flapper controllers

# FLM_STATUS response byte 1: Flapper Master Operation (current action):
FLM_OP_IDLE =        0
FLM_OP_OPENA =       1 # Open Sequence Start -- open group 'a' -- flappers 1,3
FLM_OP_WAITOPENA =   2 # waiting to complete open group a
FLM_OP_OPENB =       3 # open group 'b' flappers 2, 4
FLM_OP_WAITOPENB =   4 # waiting to complete open group b
FLM_OP_PAUSEOPENED = 5 # Open Sequence Finish
FLM_OP_CLOSEB =      6 # Close Sequence Start - close group 'b'
FLM_OP_WAITCLOSEB =  7 # waiting to complete close group b
FLM_OP_CLOSEA =      8 # close group 'a'
FLM_OP_WAITCLOSEA =  9 # waiting to  complete close group a
FLM_OP_PAUSECLOSED = 10 # Close Sequence Finish
FLM_OP_STOP =        11 # stop all flappers
FLM_OP_RECOVER =     12 # command to stop and reset all flappers
FLM_OP_ERRLOCK =     13 # flappers in locked state -- one or more flappers in fault

# FLM_STATUS response bytes 2, 3, 4, 5 contain the FLX_STATUS and FLX_OP values for
# flappers 1, 2, 3, and 4 respesctively. The high nibble is the Op State,
# and the low nibble is the Status. In other words, they are encoded as:
#   byte = ((Op & 0xf) << 4) || (Status & 0xf)
# if BYTE = 0xFF, the the data was not updated -- e.g. this is not a valid response for the flapper controller

# Lower half of each status byte
FLX_STATUS_BETWEEN = 0 # flapper is between sensors
FLX_STATUS_CLOSED =  1 # flapper is closed (close sensor active)
FLX_STATUS_OPEN =    2 # flapper is open (close sensor active)
FLX_STATUS_ERR =     3 # flapper error -- both sensors active

# Upper half of each status byte
FLX_OP_IDLE    =     0x0 # flapper is idle
FLX_OP_OPENING =     0x1 # flapper processed open command and started open sequence
FLX_OP_CLOSING =     0x2 # flapper processed close command and started close sequence
FLX_OP_FAULT   =     0xe # flapper faulted -- stops monitor operations until reset is processed
FLX_OP_TIMEOUT =     0xf # flapper open/close timeout


# Mapping from FLM_STATE value to a printable string
FLM_STATE_TO_STRING = {
    FLM_STATE_UNKNOWN: "unknown",
    FLM_STATE_OPEN: "open",
    FLM_STATE_CLOSED: "closed",
    FLM_STATE_BETWEEN: "between",
    FLM_STATE_ERR: "error"
}

# Mapping from FLM_OP value to a printable string
FLM_OP_TO_STRING = {
    FLM_OP_IDLE: "idle",
    FLM_OP_OPENA: "open_a",
    FLM_OP_WAITOPENA: "opening_a",
    FLM_OP_OPENB: "open_b",
    FLM_OP_WAITOPENB: "opening_b",
    FLM_OP_PAUSEOPENED: "open_finished",
    FLM_OP_CLOSEB: "close_b",
    FLM_OP_WAITCLOSEB: "closing_b",
    FLM_OP_CLOSEA: "close_a",
    FLM_OP_WAITCLOSEA: "closing_a",
    FLM_OP_PAUSECLOSED: "close_finished",
    FLM_OP_STOP: "stopped",
    FLM_OP_RECOVER: "recovering",
    FLM_OP_ERRLOCK: "error_locked"
}

# Mapping from FLX_STATUS value to a printable string
FLX_STATUS_TO_STRING = {
    FLX_STATUS_BETWEEN: "between",
    FLX_STATUS_CLOSED: "closed",
    FLX_STATUS_OPEN: "open",
    FLX_STATUS_ERR: "error"
}

# Mapping from FLX_OP value to a printable string
FLX_OP_TO_STRING = {
    FLX_OP_IDLE: "idle",
    FLX_OP_OPENING: "opening",
    FLX_OP_CLOSING: "closing",
    FLX_OP_FAULT: "fault",
    FLX_OP_TIMEOUT: "timeout"
}
        

# Reference to an AuxSession representing the connected shutter controller
shutter = None

def connect_serial(serial_port):
    global shutter
    shutter = auxlib.AuxSession()
    shutter.openSerial(serial_port)
    shutter.debug = DEBUG_MESSAGES

def connect_tcp(host, tcpPort):
    global shutter
    shutter = auxlib.AuxSession()
    shutter.openTcp(host, tcpPort)
    shutter.debug = DEBUG_MESSAGES

def connect_default():
    if CONNECT_VIA_TCP:
        connect_tcp(DEFAULT_TCP_HOST, DEFAULT_TCP_PORT)
    else:
        connect_serial(DEFAULT_SERIAL_PORT)

def open():
    shutter.sendReceive(ADDR_SHUTTER, FLM_OPEN)

def close():
    shutter.sendReceive(ADDR_SHUTTER, FLM_CLOSE)

def stop():
    shutter.sendReceive(ADDR_SHUTTER, FLM_STOP)

def reset():
    shutter.sendReceive(ADDR_SHUTTER, FLM_RESET)

def status():
    response = shutter.sendReceive(ADDR_SHUTTER, FLM_STATUS)
    master_state = FLM_STATE_TO_STRING[response.data[0]]
    master_operation = FLM_OP_TO_STRING[response.data[1]]
    shutter_limits = []
    shutter_operation = []
    for status in response.data[2:6]:
        operation = FLX_OP_TO_STRING[(status >> 4) & 0x0F]
        limits = FLX_STATUS_TO_STRING[status & 0x0F]
        shutter_limits.append(limits)
        shutter_operation.append(operation)
    
    print("master_state = %s" % master_state)
    print("master_operation = %s" % master_operation)
    for i in range(4):
        print("shutter%d_limits = %s" % (i+1, shutter_limits[i]))
    
    for i in range(4):
        print("shutter%d_operation = %s" % (i+1, shutter_operation[i]))

    return response.data[0]

def version():
    version_bytes = shutter.sendReceive(ADDR_SHUTTER, FLM_GET_VERSION).data
    version_field3 = (version_bytes[2] << 8) | version_bytes[3]
    return "%d.%d.%d" % (version_bytes[0], version_bytes[1], version_field3)

def status_loop():
    print("master_state, master_op, lim1, lim2, lim3, lim4, op1, op2, op3, op4")
    t0 = time.time()
    while True:
        response = shutter.sendReceive(ADDR_SHUTTER, FLM_STATUS)
        master_state = FLM_STATE_TO_STRING[response.data[0]]
        master_operation = FLM_OP_TO_STRING[response.data[1]]
        shutter_limits = []
        shutter_operation = []
        for status in response.data[2:6]:
            operation = FLX_OP_TO_STRING[(status >> 4) & 0x0F]
            limits = FLX_STATUS_TO_STRING[status & 0x0F]
            shutter_limits.append(limits)
            shutter_operation.append(operation)
        
        elapsed_sec = time.time() - t0
        print("%.2f" % elapsed_sec,
              master_state, 
              master_operation, 
              shutter_limits[0],
              shutter_limits[1],
              shutter_limits[2],
              shutter_limits[3],
              shutter_operation[0],
              shutter_operation[1],
              shutter_operation[2],
              shutter_operation[3]
              )
        time.sleep(0.2)

def usage():
    print("Usage:")
    print("  pwshutter.py [command] [status_interval]")
    print()
    print("command can be one of:")
    print("  open")
    print("  close")
    print("  stop")
    print("  reset")
    print("  status")
    print("  status_loop")
    print("  version")
    print()
    print("If status_interval is specified, then status will be repeatedly reported")
    print("with the specified period, in seconds, after sending the command.")
    print("If not specified, the program will exit immediately after sending the command.")

def main():
    if len(sys.argv) < 2:
        usage()
        return
    
    connect_default()

    command = sys.argv[1]

    if command == "open":
        open()
    
    elif command == "close":
        close()
    
    elif command == "stop":
        stop()

    elif command == "reset":
        reset()

    elif command == "status":
        status()

    elif command == "status_loop":
        status_loop()

    elif command == "version":
        print(version())

    else:
        print("Unrecognized command:", command)
        usage()
        return

    if len(sys.argv) > 2:
        loop_interval = float(sys.argv[2])
        while True:
            print()
            status()
            time.sleep(loop_interval)

if __name__ == "__main__":
    main()
